@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2002, 2003 Philip Blundell <philb@gnu.org>
+ * Copyright (C) 2002, 2003, 2006 Philip Blundell <philb@gnu.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -43,6 +43,7 @@
 #include "main.h"
 #include "sdp.h"
 #include "dbus.h"
+#include "progress.h"
 
 #include "dun.h"
 #include "lap.h"
@@ -81,6 +82,7 @@ static GtkWidget *menu_radio_on, *menu_radio_off;
 static GtkWidget *menu_devices;
 static GtkWidget *devices_window;
 static GtkWidget *iconlist;
+static GtkWidget *radio_on_progress;
 
 static GSList *devices;
 
@@ -90,45 +92,6 @@ GSList *service_desc_list;
 
 extern void bluez_pin_dbus_server_run (void);
 
-GtkWidget *
-bt_progress_dialog (gchar *text, GdkPixbuf *pixbuf)
-{
-  GtkWidget *window;
-  GtkWidget *label;
-  GtkWidget *image;
-  GtkWidget *hbox;
-
-  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
-  hbox = gtk_hbox_new (FALSE, 0);
-  image = gtk_image_new_from_pixbuf (pixbuf);
-  label = gtk_label_new (text);
-
-  gtk_window_set_type_hint (GTK_WINDOW (window), GDK_WINDOW_TYPE_HINT_DIALOG);
-  
-  gtk_container_set_border_width (GTK_CONTAINER (hbox),
-					gpe_get_border ());
-	
-  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
-
-  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
-
-  gtk_container_add (GTK_CONTAINER (window), hbox);
-
-  g_object_set_data (G_OBJECT (window), "label", label);
-
-  return window;
-}
-
-void
-bt_progress_dialog_update (GtkWidget *w, gchar *new_text)
-{
-  GtkWidget *label;
-
-  label = GTK_WIDGET (g_object_get_data (G_OBJECT (w), "label"));
-  gtk_label_set_text (GTK_LABEL (label), new_text);
-}
-
 void
 set_image(int sx, int sy)
 {
@@ -172,6 +135,28 @@ fork_hciattach (void)
   return 0;
 }
 
+#define RADIO_ON_POLL_TIME	500
+
+static void
+check_radio_startup (guint id)
+{
+  int dd;
+
+  dd = hci_open_dev (0);
+  if (dd != -1)
+    {
+      hci_close_dev (dd);
+
+      if (radio_on_progress)
+	{
+	  gtk_widget_destroy (radio_on_progress);
+	  radio_on_progress = NULL;
+	}
+    }
+  else
+    g_timeout_add (RADIO_ON_POLL_TIME, (GSourceFunc) check_radio_startup, NULL);
+}
+
 static void
 radio_on (void)
 {
@@ -190,13 +175,25 @@ radio_on (void)
   gtk_widget_show (menu_radio_off);
   gtk_widget_set_sensitive (menu_devices, TRUE);
 
-  radio_is_on = TRUE;
   sigemptyset (&sigs);
   sigaddset (&sigs, SIGCHLD);
   sigprocmask (SIG_BLOCK, &sigs, NULL);
   hciattach_pid = fork_hciattach ();
   sigprocmask (SIG_UNBLOCK, &sigs, NULL);
-  set_image(0, 0);
+
+  if (hciattach_pid == 0)
+    {
+      gpe_perror_box (_("Couldn't exec " HCIATTACH));
+      return;
+    }
+
+  radio_on_progress = bt_progress_dialog (_("Energising radio"), gpe_find_icon ("bt-logo"));
+  gtk_widget_show_all (radio_on_progress);
+
+  g_timeout_add (RADIO_ON_POLL_TIME, (GSourceFunc) check_radio_startup, NULL);
+
+  radio_is_on = TRUE;
+  set_image (0, 0);
 }
 
 static void
@@ -427,8 +424,7 @@ run_scan (gpointer data)
 	  gdk_flush ();
 	  gdk_threads_leave ();
 
-	  bd->sdp = TRUE;
-	  sdp_browse_device (bd, PUBLIC_BROWSE_GROUP);
+	  bd->sdp = sdp_browse_device (bd, PUBLIC_BROWSE_GROUP);
 
 	  gdk_threads_enter ();
 	}
@@ -478,35 +474,86 @@ show_devices (void)
     gpe_perror_box (_("Unable to scan for devices"));
 }
 
-static void
-sigchld_handler (int sig)
+static gboolean sigchld_signalled;
+
+static void 
+sigchld_handler (int signo)
+{
+  sigchld_signalled = TRUE;
+}
+
+static gboolean
+sigchld_source_dispatch (GSource *source, GSourceFunc callback, gpointer user_data)
 {
   int status;
-  pid_t p = waitpid (0, &status, WNOHANG);
+  pid_t p;
 
-  if (p == hciattach_pid)
+  sigchld_signalled = FALSE;
+  
+  do
     {
-      if (WIFSIGNALED (status) && (WTERMSIG (status) == SIGHUP))
+      p = waitpid (0, &status, WNOHANG);
+      if (p < 0)
+	perror ("waitpid");
+
+      if (p == hciattach_pid)
 	{
-	  /* restart hciattach after hangup */
-	  hciattach_pid = fork_hciattach ();
-	}
-      else
-	{
-	  hciattach_pid = 0;
-	  if (radio_is_on)
+	  if (WIFSIGNALED (status) && (WTERMSIG (status) == SIGHUP))
 	    {
-	      gpe_error_box_nonblocking (_("hciattach died unexpectedly"));
-	      radio_off ();
+	      /* restart hciattach after hangup */
+	      hciattach_pid = fork_hciattach ();
+	    }
+	  else
+	    {
+	      hciattach_pid = 0;
+	      if (radio_is_on)
+		{
+		  gpe_error_box_nonblocking (_("hciattach died unexpectedly"));
+		  radio_off ();
+		}
+	      else
+		{
+		  if (radio_on_progress)
+		    {
+		      gtk_widget_destroy (radio_on_progress);
+		      radio_on_progress = NULL;
+		    }
+		  
+		  gpe_perror_box_nonblocking (_("Radio startup failed"));
+		}
 	    }
 	}
-    }
-  else if (p > 0)
-    {
-      fprintf (stderr, "unknown pid %d exited\n", p);
-    }
+      else if (p > 0)
+	{
+	  fprintf (stderr, "unknown pid %d exited\n", p);
+	}
+    } while (p > 0);
+
+  return TRUE;
 }
 
+static gboolean
+sigchld_source_prepare (GSource *source, gint *timeout)
+{
+  *timeout = -1;
+  return sigchld_signalled;
+}
+
+static gboolean
+sigchld_source_check (GSource *source)
+{
+  return sigchld_signalled;
+}
+
+static GSourceFuncs
+sigchld_source_funcs = 
+  {
+    sigchld_source_prepare,
+    sigchld_source_check,
+    sigchld_source_dispatch,
+    NULL
+  };
+
 static void
 clicked (GtkWidget *w, GdkEventButton *ev)
 {
@@ -543,6 +590,7 @@ main (int argc, char *argv[])
   GtkWidget *menu_remove;
   GtkTooltips *tooltips;
   int dd;
+  GSource *sigchld_source;
 
   g_thread_init (NULL);
   gdk_threads_init ();
@@ -571,6 +619,8 @@ main (int argc, char *argv[])
   gtk_window_set_title (GTK_WINDOW (window), _("Bluetooth control"));
 
   signal (SIGCHLD, sigchld_handler);
+  sigchld_source = g_source_new (&sigchld_source_funcs, sizeof (GSource));
+  g_source_attach (sigchld_source, NULL);
 
   menu = gtk_menu_new ();
   menu_radio_on = gtk_menu_item_new_with_label (_("Switch radio on"));
